! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

!******************************************
! Test program for omdetect
!******************************************
module utilities_mod

use dal
use cal
use omsourceparam_mod  ! In mssllib defines Struct SourceParameters
use omutility_mod
use param

contains

!*****************************************************
!
!*****************************************************
real(single) function gasdev() result(output)

  integer, save :: iset = 0
  real :: fac, rsq, v1, v2, ranNumber
  real, save :: gset

  if(iset .eq. 0)then
     loop1: do
!---Test VNY 2.11.2001
        ranNumber=0.0

        call random_number(ranNumber)
        v1 = 2.0 * ranNumber - 1.0

        call random_number(ranNumber)
        v2 = 2.0 * ranNumber - 1.0
        rsq = v1 ** 2 + v2 ** 2
        if(rsq .ge. 1.0 .or. rsq .lt. 1e-6) then
           continue
        else
           exit loop1
        endif
     end do loop1
     fac = sqrt(-2.0 * log(rsq) / rsq)
     gset = v1 * fac
     output = v2 * fac
     iset=1
  else
     output = gset
     iset = 0
  endif
  return
end function gasdev

!**********************************************************
!  Gives the OM count-rate for the given
! star type and magnitude
!**********************************************************
real(single) Function OmCountRate(filterNo, starType, magnitude)
    integer(int8), intent(in) :: filterNo, starType
    real(single), intent(in) :: magnitude

    real(single), parameter :: rate(6, 7) = reshape(source = &
                 (/1671.0, 1659.0, 1642.0, 1611.0, 1602.0, 1590.0, 1556.0, &
                 7248.0, 5170.0, 4033.0, 2747.0, 2174.0, 2054.0, 1136.0, & 
                 9662.0, 2118.0, 1476.0, 930.0, 522.0, 403.0, 89.4, &
                 6396.0, 809.0, 378.0, 194.0, 90.3, 61.6, 10.6, &
                 2438.0, 189.0, 37.6, 7.36, 2.2, 0.47, 0.058, &
                 1100.0, 79.4, 13.4, 2.47, 0.93, 0.46, 0.21/), &
                 shape = (/6, 7/))



    ! The count rate for the given magnitude
    OmCountRate = 0.0001 * rate(filterNo, 1) * &
        (10.0 ** (0.4 * (20.0 - magnitude)))
   
end function OmCountRate  
!**********************************************************
! Gives the OM magnitude for the given star type and magnitude
!**********************************************************
real(single) Function OmMagnitude(filterNo, starType, countRate)
    integer(int8), intent(in) :: filterNo, starType
    real(single), intent(in) :: countRate
    real(single) :: omCounts
    real(single), parameter :: OmCountRate(6, 7) = reshape(source = &
                 (/1671.0, 1659.0, 1642.0, 1611.0, 1602.0, 1590.0, 1556.0, &
                 7248.0, 5170.0, 4033.0, 2747.0, 2174.0, 2054.0, 1136.0, & 
                 9662.0, 2118.0, 1476.0, 930.0, 522.0, 403.0, 89.4, &
                 6396.0, 809.0, 378.0, 194.0, 90.3, 61.6, 10.6, &
                 2438.0, 189.0, 37.6, 7.36, 2.2, 0.47, 0.058, &
                 1100.0, 79.4, 13.4, 2.47, 0.93, 0.46, 0.21/), &
                 shape = (/6, 7/))



    ! The count rate for the given magnitude
    omCounts = 0.0001 * OmCountRate(filterNo, starType)
   
    OmMagnitude = 20.0 - 2.5 * log10(countRate / omCounts) 

end function OmMagnitude  




!********************************************
!
!********************************************
real(single) Function GetMagnitude(magMax)
    real(single), intent(in) :: magMax

    real(single) :: r1, r2, r3
    real(single) :: magMin = 13.0, magRange = 7.0, alpha = 5.0
    real(single) :: r
    real(single) :: rMax 

    magMin = realParameter("minmag")
  
    magRange = magMax - magMin

    rMax =  (magMax / magMin) ** alpha 
    
    do 
        call random_number(r1)
        r1 = magMin + r1 * magRange
        r2 = (r1 / magMin) ** alpha 
        call random_number(r3)
        r3 = r3 * rMax
        if (r3 < r2) exit   
         

    end do

    GetMagnitude = r1 * exp(-(r1 - 10.0) / 10.0)
    GetMagnitude = max(magMin, r1)
    GetMagnitude = min(magMax, r1)
    

end function GetMagnitude


real(single) function SignalToNoise(sourceCountRate, backgroundCountRate)
    real(single), intent(in) :: sourceCountRate, backgroundCountRate    

    SignalToNoise = sourceCountRate / sqrt(sourceCountRate + backgroundCountRate)

end function SignalToNoise


!*********************************************************************************
! subroutine loadSources
! Info: loads sources found in dataset(setName), and returns a newly created vector
!       of sources.
! 
!********************************************************************************
subroutine loadSources(sources, setName, extensionFlag, confusionFlag, magLimit)

    type(SourceParameters), pointer    :: sources(:)
    character(*), intent(in) :: setName
    integer(int8), intent(in) :: extensionFlag, confusionFlag
    real(single), intent(out) :: magLimit
    type(SourceParameters), pointer    :: tempSources(:) => null()
    type(ColumnT) :: columnH 
!    real(single), intent(out) :: background
  
    ! locals
    integer(int32)        :: nSources
  
    type(DataSetT)        ::  set
    type(TableT)          :: inTableH
    real(single)     :: innerBackgroundRadius
    real(single)     :: outerBackgroundRadius
    real(single)     :: exposure, crLimit

    integer i, count         
    integer(int8) :: one = 1  


      ! --- don't bother if none available
    if (trim(setName) == "" ) return

    ! Read in the OM OSW Source List file.
    set = dataSet(setName, read)


write(*, *) "Loadsources loading sources from ", trim(setName)      

    if (.not. hasBlock(set, "SRCLIST")) then
    

        ! Incompatible file - can't continue
        write (errstr, *) "Could not find any compatibile blocks in set ", trim(setName)
        call fatal("BadFileType")

    else
        ! Get a handle on the source list binary table extension.
        inTableH = table(set, "SRCLIST")

    end if

    if (associated(sources)) deallocate(sources)
    nullify(sources)     
    
    
    innerBackgroundRadius =  GetReal32Attribute(set, "INNBCKAN")
    outerBackgroundRadius =  GetReal32Attribute(set, "OUTBCKAN")
    exposure =  GetReal32Attribute(set, "EXPOSURE")


    crLimit =  GetReal32Attribute(set, "CRLIMIT")
    magLimit = OmMagnitude(one, one, crLimit)
  

!write(*, *) "Line 212"
    write(errstr, '(a, f6.2)') "Omdetect detection limiting magnitude=",  magLimit      
    call message(AppMsg, Verbose)
 




    !***********************************************
    ! Get the background
    !***********************************************
!    background =  GetReal32Attribute(set, "CRLIMIT")
!    background = ((background * exposure / 2.0) ** 2.0) / (PI * 36.0) 
!    write(*, *) "Load sources background=", background, " exposure=", exposure       
                 
    ! Get the number of rows in the source list binary table extension.
    nSources = numberOfRows(inTableH)

    if (nSources == 0) then
      

        call release(inTableH)
        call release(set)
        write(*, *) "loadsources returning at line 235- no sources"
        return

    end if


write(*, *) "Line 240 no of rows=", numberOfRows(intableh)
   
    allocate(tempSources(1 : nSources))
   
 
    ! Get the data for each source from the source list file

  
    call LoadReal32Column(inTableH, "XPOS",  tempSources%xPos)
 
  
    call LoadReal32Column(inTableH, "YPOS",  tempSources%yPos)
    call LoadReal32Column(inTableH, "POSERR",  tempSources%poserr)


 
  
    call LoadReal32Column(inTableH, "FWHM_MAJ",  tempSources%semiA)
 
    call LoadReal32Column(inTableH, "FWHM_MAJ_ERR",  tempSources%semiAErr)
 

 
  
    call LoadReal32Column(inTableH, "FWHM_MIN",  tempSources%semiB)
    call LoadReal32Column(inTableH, "FWHM_MIN_ERR",  tempSources%semiBErr)

 
  
    call LoadReal32Column(inTableH, "PA", tempSources%theta)
   call LoadReal32Column(inTableH, "PA_ERR",  tempSources%thetaErr)
 
  
    call LoadReal32Column(inTableH, "RATE", tempSources%rawCountRate)
 
  
    call LoadReal32Column(inTableH, "RATE_ERR", tempSources%rawCountRateErr)
 

    call LoadReal32Column(inTableH, "CORR_RATE", tempSources%corrCountRate)
  
  
    call LoadReal32Column(inTableH, "CORR_RATE_ERR", tempSources%corrCountRateErr)
 
  
    call LoadReal32Column(inTableH, "COINCLOSS_CORR", tempSources%coincidenceLossCorr)
  
    call LoadReal32Column(inTableH, "PSF1_CORR", tempSources%psf1Correction)
    
  
    call LoadReal32Column(inTableH, "PSF2_CORR", tempSources%psf2Correction)
 

    call LoadReal32Column(inTableH, "BACKGROUND_RATE",  tempSources%rawBackgroundCountRate)
  
    call LoadReal32Column(inTableH, "CORR_BACKGROUND_RATE",  tempSources%corrbackgroundCountRate)
  

    call LoadReal32Column(inTableH, "MAG", tempSources%magnitude)
  
  
    call LoadReal32Column(inTableH, "MAGERR", tempSources%magnitudeErr)
  
  
    call LoadReal32Column(inTableH, "SIGNIFICANCE", tempSources%significance)
    
  
    call LoadReal32Column(inTableH, "APERTURE", tempSources%apertureRadius)


    call LoadInt8Column(inTableH, "EFLAG", tempSources%extensionFlag)
    
  
  

    ! Get the RAS and decs if they exist
    call LoadReal64Column(inTableH, "RA", tempSources%rightAscension)

  
    call LoadReal64Column(inTableH, "DEC", tempSources%declination)
  
  
    columnH = column(inTableH, "QFLAG")

    if (dataType(columnh) .eq. integer16) then
        call LoadInt16Column(inTableH, "QFLAG", tempSources%qualityFlag)
    else
        call LoadInt8Column(inTableH, "QFLAG", tempSources%confusionFlag)
        tempSources%qualityFlag =  int(tempSources%confusionFlag, int16)  
    end if
    call release(columnH)


    call LoadInt8Column(inTableH, "CFLAG", tempSources%confusionFlag)


!write(*, *) "utilities_mod.f90 Line 324 load sources loaded confusion flags"
!do i = 1, nsources
!write(*, *) i, " confusion flag=", tempsources(i)%confusionFlag
!end do
 

    call release(inTableH)
    call release(set)


!write(*, *) "lINE 324 EXTENSION flag=", extensionFlag, " confusionflag=", confusionFlag

    if (extensionFlag .lt. 0 .and. confusionFlag .lt. 0) then
        allocate(sources(1 : nSources))

write(*, *) "Line 352, nsources=", nsources, " size(sources)=", size(sources)
        sources = tempSources

    else   ! Need to filter source-list
 
!write(*, *) "Line 362"
        count = 0
        do i = 1, nSources

!write(*, *) i, " confusion flag=", tempsources(i)%confusionFlag
           if ( (extensionFlag .eq. tempSources(i)%extensionFlag .or. &
                extensionFlag .lt. 0)  .and. &
                     (tempSources(i)%confusionFlag .le. confusionFlag &
             .or. confusionFlag .eq. -1_int8)) then
                  
               count = count + 1
           end if
      
        end do

!write(*, *) "Line 374 count=", count


    
        allocate(sources(1 : count))
        count = 0
        do i = 1, nSources

            if ( (extensionFlag .eq. tempSources(i)%extensionFlag .or. &
                  extensionFlag .lt. 0) .and. &
                     (tempSources(i)%confusionFlag .le. confusionFlag &
                     .or. confusionFlag .eq. -1_int8)) then
         
            
                count = count + 1
                sources(count) = tempSources(i)
       
            end if


        end do
    end if

!write(*, *) "utilities_mod Line 397 count=", count, " extensionflag=", extensionFlag, " confusionflag=", confusionFlag
    sources%innerBckRadius = innerBackgroundRadius

    sources%outerBckRadius = outerBackgroundRadius
    sources%exposure = exposure   

    deallocate(tempSources)
    nullify(tempSources)

end subroutine loadSources

!***********************************************
!
!***********************************************
subroutine AddMagnitudeColumnsToTable(srcListFile, source)     
        
    character(*), intent(in) :: srcListFile
    type(SourceParameters) :: source(:) 
    type(DataSetT) ::  set
    type(TableT) :: tableSet
    set = DataSet(srcListFile, modify)
    tableSet = Table(set, "SRCLIST", modify)

    call AddReal32Column(tableSet, source%magnitude, size(source), "MAG", &
        "mag", "Instrumental magnitude", -999.9)

    call AddReal32Column(tableSet, source%magnitudeErr, size(source), "MAGERR", &
        "mag", "Instrumental magnitude error", -999.9)


    call release(tableSet)
    call release(set)

end subroutine  AddMagnitudeColumnsToTable


end module utilities_mod













